/*
 Logic Observer - a CPLD Based Logic Analyser
 Version: 1.0 Rev 2
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 12/12/2013
 Last Modified: 26/04/2015
 
 An 8 input CPLD based Logic Analyser with a selectable sample rate up to 50 MSPS with up to 131Kbit sample depth per input, 
 controllable via USB with a GUI or command line interface and data format readable by Sigrok PulseView.
 
 This project uses an Altera 64 macro cell (EPM3064ATC44) to increment the Cypress 1Mbit SRAM (CY7C1019DV33) address lines
 and an Atmel ATmega48/88/168/328 with V-USB controls the trigger settings.
 
 Remember to change your fuse bits so the ATmega uses a 20MHz crystal
 avrdude -p atmega48 -c usbtiny -U lfuse:w:0xde:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m
 
 */

// ATmega48/88/168/328 QFP Pin Map
//
// 							PB7 XTAL											PD7 Read byte 7
// PC6 Reset				PB6 XTAL			 								PD6 Clock output
// PC5 Read byte 5		PB5 SCK / Capture LED						PD5 Anlg Switch Select
// PC4 Read byte 4		PB4 MISO											PD4 
// PC3 Read byte 3		PB3 MOSI											PD3 USB D-
// PC2 Read byte 2		PB2 CPLD Reading Pin / Buffers OE		PD2 USB D+
// PC1 Read byte 1		PB1 CPLD Reading Next Address				PD1 CPLD Start Capture
// PC0 Read byte 0		PB0 Read byte 6								PD0 CPLD Reset

#define F_CPU 20000000 // If this will be changed, modify it in usbconfig.h too

#define cpldresetPin PD0
#define startcapturePin PD1
#define analogswitchPin PD5
#define nextaddressPin PB1
#define readingPin PB2
#define captureLED PB5

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include <string.h>
#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "setup.c"

uint8_t captureComplete = false;
uint8_t cancelCapture = false;
uint32_t addressCount = 0;
uint32_t addressEnd = 131064; // SRAM end (approx)
volatile uint16_t triggerDelayCounter = 0;
volatile uint8_t performUsbPoll = 0;
volatile uint8_t wait_trigger = 0;
uint8_t pin_reading_before = 0;

// V-USB Transfer variables
#define USB_TRANSFER 1
#define USB_SETTINGS 2
#define USB_STATUS 3
#define USB_CANCEL 4
static uchar settings[9];
static uchar replyBuf[254];
static uchar *nextDigit;
static uchar dataReceived = 0, dataLength = 0; // for USB_DATA_IN

// This gets called when custom control message is received
USB_PUBLIC uchar usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = (void *)data; // cast data to correct type
	
	if (rq->bRequest == USB_TRANSFER && addressCount < addressEnd) { // custom command is in the bRequest field
		nextDigit = &replyBuf[-1];
		
		// Loop 254 times to send this data over USB
		PORTB |= (1<<captureLED); // Show activity
		for (uint8_t a = 0; a < 254; a++) {
			PORTB |= (1<<nextaddressPin); // Tell the CPLD we want the next address
			asm ("nop"); // Wait a little bit for SRAM to show proper reading
			asm ("nop");
			asm ("nop");
			asm ("nop");
			asm ("nop");
			
			// Read and combine PD7, PB0 & PC5-0
			uint8_t pinb_reading = ((PINB & 0x01) << 6);
			uint8_t reading = ((PIND & 0x80) | (PINC & 0x3F)) | pinb_reading;
			*++nextDigit = reading;
			
			PORTB &= ~(1<<nextaddressPin); 
			addressCount++;
		}
		PORTB &= ~(1<<captureLED);
		
		usbMsgPtr = replyBuf;
		return sizeof(replyBuf);
	}
	else if (rq->bRequest == USB_SETTINGS) { // Re-program the settings
		dataLength = (uchar)rq->wLength.word;
		dataReceived = 0;
		
		if (dataLength > sizeof(replyBuf)) { // Limit to buffer size
			dataLength = sizeof(replyBuf);
		}
		
		// Reset address count so data can be re-read again if needed
		addressCount = 0;
		
		return USB_NO_MSG; // usbFunctionWrite will be called now
	}
	else if (rq->bRequest == USB_STATUS) { // Status of the capture
		nextDigit = &replyBuf[-1];
		
		if (captureComplete == true) {
			*++nextDigit = '1';
		}
		else {
			*++nextDigit = '0';
		}
		
		*++nextDigit = '\0';
		usbMsgPtr = replyBuf;
		return sizeof(replyBuf);
	}
	else if (rq->bRequest == USB_CANCEL) { // Cancel capture
		nextDigit = &replyBuf[-1];
		
		if (captureComplete == false) {
			cancelCapture = true;
		}
		
		*++nextDigit = '1';
		*++nextDigit = '\0';
		usbMsgPtr = replyBuf;
		return sizeof(replyBuf);
	}
	
	return 0;
}

// This gets called when data is sent from PC to the device
USB_PUBLIC uchar usbFunctionWrite(uchar *data, uchar len) {
	// Store in settings array
	for (uint8_t i = 0; dataReceived < dataLength && i < len; i++, dataReceived++) {
		settings[i] = data[i];
	}
	
	// Set sample depth
	addressEnd = samples_size[(settings[SAMPLES])];
	
	// Reading mode, switch to 10MHz clock for reading
	if (settings[CAPTURE] == false) {
		PORTD |= (1<<analogswitchPin);
		OCR0A = 0;
		_delay_ms(5);
	}
	
	// Reset CPLD (start address is always kept)
	PORTD &= ~(1<<cpldresetPin);
	_delay_ms(1);
	PORTD |= (1<<cpldresetPin);
	
	return (dataReceived == dataLength); // 1 if we received it all, 0 if not
}

// Start capturing and wait for a trigger
void start_capture(void) {
	PORTD &= ~(1<<cpldresetPin); // Reset CPLD
	captureComplete = false; // Reset vars
	
	// Change analog switch to 50MHz clock if needed
	if (settings[SAMPLE_RATE] == 0x50) { 
		PORTD &= ~(1<<analogswitchPin);
	}
	else { // Or to AVR timer clock
		OCR0A = settings[SAMPLE_RATE];
		PORTD |= (1<<analogswitchPin);
	}
	
	// Tell CPLD to start capturing data
	PORTB &= ~(1<<readingPin);
	PORTD |= (1<<cpldresetPin);
	
	_delay_ms(25); // Wait a little for clock to stabilise / Clear the start of SRAM
	
	TCNT1 = 0; // Reset usbPoll timer 
	sbi(TCCR1B, CS11); // Start timer - 8 prescaler (26ms)
	PORTB |= (1<<captureLED); // Turn on capture LED
	wait_trigger = 0; // Reset trigger count
	
	// Read pins before pin interrupt to see if they are at the state right away and capture their state
	uint8_t pinb_reading = ((PINB & 0x01) << 6);
	uint8_t pin_reading = ((PIND & 0x80) | (PINC & 0x3F)) | pinb_reading;
	pin_reading_before = pin_reading; // Store the state
	if (settings[TRIGGER_LEVEL] == HIGH) { // Waiting for high reading
		if (pin_reading & settings[TRIGGER_PINS]) {
			wait_trigger++;
		}
	}
	else if ((pin_reading ^ settings[TRIGGER_PINS]) & settings[TRIGGER_PINS]) { // Low reading
		wait_trigger++;
	}
	
	// Turn on pin change interrupt and wait for more triggers
	if (wait_trigger < settings[TRIGGER_COUNT]) { // Waiting for pin change interrupt
		// Turn on pin change interrupt for the selected pins
		if (settings[TRIGGER_PINS] & 0x3F) { // Pin 1-6
			sbi(PCICR, PCIE1);
			PCMSK1 = settings[TRIGGER_PINS] & 0x3F;
		}
		if (settings[TRIGGER_PINS] & 0x40) { // Pin 7
			sbi(PCICR, PCIE0);
			PCMSK0 = (1<<PCINT0);
		}
		if (settings[TRIGGER_PINS] & 0x80) { // Pin 8
			sbi(PCICR, PCIE2);
			PCMSK2 = (1<<PCINT23);
		}
		
		// Wait for trigger
		while (wait_trigger < settings[TRIGGER_COUNT]) { // Waiting for pin change interrupt
			if (performUsbPoll == true) {
				usbPoll();
				performUsbPoll = false;
			}
			
			// Cancel if requested
			if (cancelCapture == true) {
				break;
			}
		}
	}
	
	if (cancelCapture != true) {
		// Trigger delay
		if (settings[TRIGGER_DELAY] >= 1) {
			triggerDelayCounter = 0;
			
			// Setup Timer 2
			TCNT2 = 0; // Reset timer counter
			sbi(TCCR2B, CS20); // Start timer - no prescaler
			
			// Wait until end
			uint16_t timerEnd = (uint16_t) settings[TRIGGER_DELAY] * (uint16_t) 78;
			while (triggerDelayCounter < timerEnd) {
				if (performUsbPoll == true) {
					usbPoll();
					performUsbPoll = false;
				}
			}
			cbi(TCCR2B, CS20); // Stop timer
		}
		
		// Trigger the CPLD to write the start address of our capture
		PORTD |= (1<<startcapturePin);
		
		// Wait for the capture (to allow for all sample rate ranges)
		if (settings[SAMPLE_RATE] == 0x50) {
			_delay_ms(10);
		}
		else if (settings[SAMPLE_RATE] <= 0x09) {
			_delay_ms(200);
		}
		else {
			_delay_ms(1500);
		}
	}
	
	PORTB &= ~(1<<captureLED);
	TCCR1B = 0; // Turn off usbPoll timer
	
	// Finish writing
	PORTB |= (1<<readingPin);
	PORTD &= ~(1<<startcapturePin);
	
	// Change analog switch to 10MHz clock
	PORTD |= (1<<analogswitchPin);
	OCR0A = 0;
	
	// Update variables
	addressCount = 0;
	captureComplete = true;
	cancelCapture = false;
}

int main (void) {
	setup();
	
	// Start USB
	startUSB();
	
	while (1) {
		usbPoll();
		
		// Start capture if set
		if (settings[CAPTURE] == true) {
			start_capture();
			settings[CAPTURE] = false;
		}
	}
	
	return 0;
}

// ~26ms overflow
ISR(TIMER1_OVF_vect) {
	performUsbPoll = 1;
}

// ~1ms overflow
ISR(TIMER2_OVF_vect) {
	triggerDelayCounter++;
}

// Pin change detected, read pins and increment trigger if trigger condition matches
ISR(PCINT0_vect) {
	uint8_t pinb_reading = ((PINB & 0x01) << 6);
	uint8_t pin_reading = ((PIND & 0x80) | (PINC & 0x3F)) | pinb_reading;
	
	if (settings[TRIGGER_LEVEL] == HIGH) { // Waiting for high reading
		uint8_t high_pins = pin_reading & settings[TRIGGER_PINS]; // Check against trigger pins to see high pins only
		if (high_pins & (~pin_reading_before)) { //	Check if pin went low before going high
			wait_trigger++;
		}
	}
	else {
		uint8_t low_pins = (pin_reading ^ settings[TRIGGER_PINS]) & settings[TRIGGER_PINS];
		if (low_pins & pin_reading_before) {//	Check if pin went high before going low
			wait_trigger++;
		}
	}
	
	pin_reading_before = pin_reading; // Store the new state
	
	// Trigger matches our trigger condition, turn off pin interrupts
	if (wait_trigger >= settings[TRIGGER_COUNT]) {
		PCICR = 0;
		PCMSK2 = 0;
		PCMSK1 = 0;
		PCMSK0 = 0;
	}
}

ISR_ALIAS(PCINT1_vect, PCINT0_vect);
ISR_ALIAS(PCINT2_vect, PCINT0_vect);
